/*
 * This file is part of hat-trie.
 *
 * Copyright (c) 2011 by Daniel C. Jones <dcjones@cs.washington.edu>
 *
 *
 * Common typedefs, etc.
 *
 */


#ifndef HATTRIE_COMMON_H
#define HATTRIE_COMMON_H

typedef unsigned long value_t;

#endif


